# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, cast

from gravitino.api.rel.types.json_serdes._helper.serdes_utils import (
    SerdesUtils as TypesSerdesUtils,
)
from gravitino.dto.rel.expressions.field_reference_dto import FieldReferenceDTO
from gravitino.dto.rel.expressions.func_expression_dto import FuncExpressionDTO
from gravitino.dto.rel.expressions.function_arg import FunctionArg
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.dto.rel.expressions.unparsed_expression_dto import UnparsedExpressionDTO
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class SerdesUtils(SerdesUtilsBase):
    @classmethod
    def write_function_arg(cls, arg: FunctionArg) -> Dict[str, Any]:
        arg_type = arg.arg_type()
        if arg_type not in FunctionArg.ArgType:
            raise ValueError(f"Unknown function argument type: {arg_type}")

        arg_data = {cls.EXPRESSION_TYPE: arg_type.name.lower()}
        if arg_type is FunctionArg.ArgType.LITERAL:
            expression = cast(LiteralDTO, arg)
            arg_data[cls.DATA_TYPE] = TypesSerdesUtils.write_data_type(
                data_type=expression.data_type()
            )
            arg_data[cls.LITERAL_VALUE] = expression.value()

        if arg_type is FunctionArg.ArgType.FIELD:
            arg_data[cls.FIELD_NAME] = cast(FieldReferenceDTO, arg).field_name()

        if arg_type is FunctionArg.ArgType.FUNCTION:
            expression = cast(FuncExpressionDTO, arg)
            arg_data[cls.FUNCTION_NAME] = expression.function_name()
            arg_data[cls.FUNCTION_ARGS] = [
                cls.write_function_arg(func_arg) for func_arg in expression.args()
            ]

        if arg_type is FunctionArg.ArgType.UNPARSED:
            expression = cast(UnparsedExpressionDTO, arg)
            arg_data[cls.UNPARSED_EXPRESSION] = expression.unparsed_expression()

        return arg_data

    @classmethod
    def read_function_arg(cls, data: Dict[str, Any]) -> FunctionArg:
        Precondition.check_argument(
            data is not None and isinstance(data, dict),
            f"Cannot parse function arg from invalid JSON: {data}",
        )
        Precondition.check_argument(
            data.get(cls.EXPRESSION_TYPE) is not None,
            f"Cannot parse function arg from missing type: {data}",
        )
        try:
            arg_type = FunctionArg.ArgType(data[cls.EXPRESSION_TYPE].lower())
        except ValueError as exc:
            raise IllegalArgumentException(
                f"Unknown function argument type: {data[cls.EXPRESSION_TYPE]}"
            ) from exc

        if arg_type is FunctionArg.ArgType.LITERAL:
            Precondition.check_argument(
                data.get(cls.DATA_TYPE) is not None,
                f"Cannot parse literal arg from missing data type: {data}",
            )
            Precondition.check_argument(
                data.get(cls.LITERAL_VALUE) is not None,
                f"Cannot parse literal arg from missing literal value: {data}",
            )
            return (
                LiteralDTO.builder()
                .with_data_type(
                    data_type=TypesSerdesUtils.read_data_type(data[cls.DATA_TYPE])
                )
                .with_value(value=data[cls.LITERAL_VALUE])
                .build()
            )

        if arg_type is FunctionArg.ArgType.FIELD:
            Precondition.check_argument(
                data.get(cls.FIELD_NAME) is not None,
                f"Cannot parse field reference arg from missing field name: {data}",
            )
            return (
                FieldReferenceDTO.builder()
                .with_field_name(field_name=data[cls.FIELD_NAME])
                .build()
            )

        if arg_type is FunctionArg.ArgType.FUNCTION:
            Precondition.check_argument(
                data.get(cls.FUNCTION_NAME) is not None,
                f"Cannot parse function function arg from missing function name: {data}",
            )
            Precondition.check_argument(
                data.get(cls.FUNCTION_ARGS) is not None,
                f"Cannot parse function function arg from missing function args: {data}",
            )
            args = [cls.read_function_arg(arg) for arg in data[cls.FUNCTION_ARGS]]
            return (
                FuncExpressionDTO.builder()
                .with_function_name(function_name=data[cls.FUNCTION_NAME])
                .with_function_args(function_args=args or FunctionArg.EMPTY_ARGS)
                .build()
            )

        Precondition.check_argument(
            arg_type is FunctionArg.ArgType.UNPARSED,
            f"Unknown function argument type: {arg_type}",
        )
        Precondition.check_argument(
            isinstance(data.get(cls.UNPARSED_EXPRESSION), str),
            f"Cannot parse unparsed expression from missing string field unparsedExpression: {data}",
        )
        return UnparsedExpressionDTO(data[cls.UNPARSED_EXPRESSION])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, ClassVar

from gravitino.api.metadata_object import MetadataObject


class MetadataObjectImpl(MetadataObject):
    _DOT: ClassVar[str] = "."

    def __init__(self, names: List[str], metadata_object_type: MetadataObject.Type):
        self._name = self._DOT.join(names)
        self._metadata_object_type = metadata_object_type

    def type(self) -> MetadataObject.Type:
        return self._metadata_object_type

    def name(self) -> str:
        return self._name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import Any

from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.partitions.partition import Partition


class RangePartition(Partition):
    """
    A range partition represents a result of range partitioning. For example, for range partition

    ```
    PARTITION p20200321 VALUES LESS THAN ("2020-03-22")
    ```

    its upper bound is "2020-03-22" and its lower bound is null.

    APIs that are still evolving towards becoming stable APIs, and can change from one feature
    release to another (0.5.0 to 0.6.0).
    """

    @abstractmethod
    def upper(self) -> Literal[Any]:
        """
        Returns:
            Literal[Any]: The upper bound of the partition.
        """
        pass

    @abstractmethod
    def lower(self) -> Literal[Any]:
        """
        Returns:
            Literal[Any]: The lower bound of the partition.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Optional

from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.partitions.identity_partition import IdentityPartition
from gravitino.api.rel.partitions.list_partition import ListPartition
from gravitino.api.rel.partitions.partition import Partition
from gravitino.api.rel.partitions.range_partition import RangePartition


class Partitions:
    """The helper class for partition expressions."""

    EMPTY_PARTITIONS: List[Partition] = []
    """
    An empty array of partitions
    """

    @staticmethod
    def range(
        name: str,
        upper: Literal[Any],
        lower: Literal[Any],
        properties: Optional[Dict[str, str]],
    ) -> RangePartition:
        """
        Creates a range partition.

        Args:
            name: The name of the partition.
            upper: The upper bound of the partition.
            lower: The lower bound of the partition.
            properties: The properties of the partition.

        Returns:
            The created partition.
        """
        return RangePartitionImpl(name, upper, lower, properties)

    @staticmethod
    def list(
        name: str,
        lists: List[List[Literal[Any]]],
        properties: Optional[Dict[str, str]],
    ) -> ListPartition:
        """
        Creates a list partition.

        Args:
            name: The name of the partition.
            lists: The values of the list partition.
            properties: The properties of the partition.

        Returns:
            The created partition.
        """
        return ListPartitionImpl(name, lists, properties or {})

    @staticmethod
    def identity(
        name: Optional[str],
        field_names: List[List[str]],
        values: List[Literal[Any]],
        properties: Optional[Dict[str, str]] = None,
    ) -> IdentityPartition:
        """
        Creates an identity partition.

        The `values` must correspond to the `field_names`.

        Args:
            name: The name of the partition.
            field_names: The field names of the identity partition.
            values: The value of the identity partition.
            properties: The properties of the partition.

        Returns:
            The created partition.
        """
        return IdentityPartitionImpl(name, field_names, values, properties or {})


class RangePartitionImpl(RangePartition):
    """
    Represents a result of range partitioning.
    """

    def __init__(
        self,
        name: str,
        upper: Literal[Any],
        lower: Literal[Any],
        properties: Optional[Dict[str, str]],
    ):
        self._name = name
        self._upper = upper
        self._lower = lower
        self._properties = properties

    def upper(self) -> Literal[Any]:
        """Returns the upper bound of the partition."""
        return self._upper

    def lower(self) -> Literal[Any]:
        """Returns the lower bound of the partition."""
        return self._lower

    def name(self) -> str:
        return self._name

    def properties(self) -> Dict[str, str]:
        return self._properties

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, RangePartitionImpl):
            return False
        return (
            self._name == other._name
            and self._upper == other._upper
            and self._lower == other._lower
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (self._name, self._upper, self._lower, frozenset(self._properties.items()))
        )


class ListPartitionImpl(ListPartition):
    def __init__(
        self,
        name: str,
        lists: List[List[Literal[Any]]],
        properties: Optional[Dict[str, str]],
    ):
        self._name = name
        self._lists = lists
        self._properties = properties

    def lists(self) -> List[List[Literal[Any]]]:
        """Returns the values of the list partition."""
        return self._lists

    def name(self) -> str:
        return self._name

    def properties(self) -> Dict[str, str]:
        return self._properties

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ListPartitionImpl):
            return False
        return (
            self._name == other._name
            and self._lists == other._lists
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                tuple(tuple(sublist) for sublist in self._lists),
                frozenset(self._properties.items()),
            )
        )


class IdentityPartitionImpl(IdentityPartition):
    def __init__(
        self,
        name: str,
        field_names: List[List[str]],
        values: List[Literal[Any]],
        properties: Dict[str, str],
    ):
        self._name = name
        self._field_names = field_names
        self._values = values
        self._properties = properties

    def field_names(self) -> List[List[str]]:
        """Returns the field names of the identity partition."""
        return self._field_names

    def values(self) -> List[Literal[Any]]:
        """Returns the values of the identity partition."""
        return self._values

    def name(self) -> str:
        return self._name

    def properties(self) -> Dict[str, str]:
        return self._properties

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IdentityPartitionImpl):
            return False
        return (
            self._name == other._name
            and self._field_names == other._field_names
            and self._values == other._values
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                tuple(tuple(fn) for fn in self._field_names),
                tuple(self._values),
                frozenset(self._properties.items()),
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import overload

from dataclasses_json.core import Json

from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.types.json_serdes.base import JsonSerializable
from gravitino.dto.rel.expressions.json_serdes._helper.serdes_utils import SerdesUtils


class ColumnDefaultValueSerdes(JsonSerializable[Expression]):
    """Custom JSON serializer/deserializer for Column default value."""

    @classmethod
    def serialize(cls, value: Expression) -> Json:
        if cls.is_empty(value):
            return None
        return SerdesUtils.write_function_arg(arg=value)

    @classmethod
    def deserialize(cls, data: Json) -> Expression:
        if cls.is_empty(data):
            return Column.DEFAULT_VALUE_NOT_SET
        return SerdesUtils.read_function_arg(data=data)

    @classmethod
    @overload
    def is_empty(cls, value: Expression) -> bool: ...

    @classmethod
    @overload
    def is_empty(cls, value: Json) -> bool: ...

    @classmethod
    def is_empty(cls, value):
        if isinstance(value, (Expression, list)):
            return value is None or value is Column.DEFAULT_VALUE_NOT_SET
        return value is None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List

from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.distributions.strategy import Strategy
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.named_reference import NamedReference


class DistributionImpl(Distribution):
    _strategy: Strategy
    _number: int
    _expressions: List[Expression]

    def __init__(self, strategy: Strategy, number: int, expressions: List[Expression]):
        self._strategy = strategy
        self._number = number
        self._expressions = expressions

    def strategy(self) -> Strategy:
        return self._strategy

    def number(self) -> int:
        return self._number

    def expressions(self) -> List[Expression]:
        return self._expressions

    def __str__(self) -> str:
        return f"DistributionImpl(strategy={self._strategy}, number={self._number}, expressions={self._expressions})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DistributionImpl):
            return False
        return (
            self._strategy == other.strategy()
            and self._number == other.number()
            and self._expressions == other.expressions()
        )

    def __hash__(self) -> int:
        return hash((self._strategy, self._number, tuple(self._expressions)))


class Distributions:
    NONE: Distribution = DistributionImpl(Strategy.NONE, 0, Expression.EMPTY_EXPRESSION)
    """NONE is used to indicate that there is no distribution."""
    HASH: Distribution = DistributionImpl(Strategy.HASH, 0, Expression.EMPTY_EXPRESSION)
    """List bucketing strategy hash, TODO: #1505 Separate the bucket number from the Distribution."""
    RANGE: Distribution = DistributionImpl(
        Strategy.RANGE, 0, Expression.EMPTY_EXPRESSION
    )
    """List bucketing strategy range, TODO: #1505 Separate the bucket number from the Distribution."""

    @staticmethod
    def even(number: int, *expressions: Expression) -> Distribution:
        """
        Create a distribution by evenly distributing the data across the number of buckets.

        :param number: The number of buckets.
        :param expressions: The expressions to distribute by.
        :return: The created even distribution.
        """
        return DistributionImpl(Strategy.EVEN, number, list(expressions))

    @staticmethod
    def hash(number: int, *expressions: Expression) -> Distribution:
        """
        Create a distribution by hashing the data across the number of buckets.

        :param number: The number of buckets.
        :param expressions: The expressions to distribute by.
        :return: The created hash distribution.
        """
        return DistributionImpl(Strategy.HASH, number, list(expressions))

    @staticmethod
    def of(strategy: Strategy, number: int, *expressions: Expression) -> Distribution:
        """
        Create a distribution by the given strategy.

        :param strategy: The strategy to use.
        :param number: The number of buckets.
        :param expressions: The expressions to distribute by.
        :return: The created distribution.
        """
        return DistributionImpl(strategy, number, list(expressions))

    @staticmethod
    def fields(
        strategy: Strategy, number: int, *field_names: List[str]
    ) -> Distribution:
        """
        Create a distribution on columns. Like distribute by (a) or (a, b), for complex like
        distributing by (func(a), b) or (func(a), func(b)), please use DistributionImpl.Builder to create.

        NOTE: a, b, c are column names.

        SQL syntax: distribute by hash(a, b) buckets 5
        fields(Strategy.HASH, 5, ["a"], ["b"])

        SQL syntax: distribute by hash(a, b, c) buckets 10
        fields(Strategy.HASH, 10, ["a"], ["b"], ["c"])

        SQL syntax: distribute by EVEN(a) buckets 128
        fields(Strategy.EVEN, 128, ["a"])

        :param strategy: The strategy to use.
        :param number: The number of buckets.
        :param field_names: The field names to distribute by.
        :return: The created distribution.
        """
        expressions = [NamedReference.field(name) for name in field_names]
        return Distributions.of(strategy, number, *expressions)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import Dict


class Partition(ABC):
    """
    A partition represents a result of partitioning a table. The partition can be either a
    `IdentityPartition`, `ListPartition`, or `RangePartition`. It depends on the `Table.partitioning()`.

    APIs that are still evolving towards becoming stable APIs, and can change from one feature
    release to another (0.5.0 to 0.6.0).
    """

    @abstractmethod
    def name(self) -> str:
        """
        Returns:
            str: The name of the partition.
        """
        pass

    @abstractmethod
    def properties(self) -> Dict[str, str]:
        """
        Returns:
            Dict[str, str]: The properties of the partition, such as statistics, location, etc.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import Any, List

from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.partitions.partition import Partition


class IdentityPartition(Partition):
    """
    An identity partition represents a result of identity partitioning. For example, for Hive
    partition

    ```
    PARTITION (dt='2008-08-08',country='us')
    ```

    its partition name is "dt=2008-08-08/country=us", field names are [["dt"], ["country"]] and
    values are ["2008-08-08", "us"].

    APIs that are still evolving towards becoming stable APIs, and can change from one feature
    release to another (0.5.0 to 0.6.0).
    """

    @abstractmethod
    def field_names(self) -> List[List[str]]:
        """
        Returns:
            List[List[str]]: A list of lists representing the field names of the identity partition.
        """
        pass

    @abstractmethod
    def values(self) -> List[Literal[Any]]:
        """
        Returns:
            List[Literal[Any]]: The values of the identity partition.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
from typing import Any, Dict, Union, overload

from dataclasses_json.core import Json

from gravitino.api.rel.types.type import Name, Type
from gravitino.api.rel.types.types import Types
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class SerdesUtils(SerdesUtilsBase):
    @classmethod
    def write_data_type(cls, data_type: Type) -> Union[str, Dict[str, Any]]:
        """Write Gravitino Type to JSON data. Used for Gravitino Type JSON Serialization.

        Args:
            data_type (Any): The Gravitino Type.

        Returns:
            Union[str, Dict[str, Any]]: The serialized data.
        """

        type_name = data_type.name()
        if type_name in cls.PRIMITIVE_AND_NULL_TYPES:
            return data_type.simple_string()

        if type_name is Name.STRUCT:
            return cls.write_struct_type(data_type)

        if type_name is Name.LIST:
            return cls.write_list_type(data_type)

        if type_name is Name.MAP:
            return cls.write_map_type(data_type)

        if type_name is Name.UNION:
            return cls.write_union_type(data_type)

        if type_name is Name.EXTERNAL:
            return cls.write_external_type(data_type)

        if type_name is Name.UNPARSED:
            return cls.write_unparsed_type(data_type)

        return cls.write_unparsed_type(data_type.simple_string())

    @classmethod
    def write_struct_type(cls, struct_type: Types.StructType) -> Dict[str, Any]:
        struct_data = {
            cls.TYPE: cls.STRUCT,
            cls.FIELDS: [
                cls.write_struct_field(field) for field in struct_type.fields()
            ],
        }
        return struct_data

    @classmethod
    def write_struct_field(cls, struct_field: Types.StructType.Field) -> Dict[str, Any]:
        struct_field_data = {
            cls.STRUCT_FIELD_NAME: struct_field.name(),
            cls.TYPE: cls.write_data_type(struct_field.type()),
            cls.STRUCT_FIELD_NULLABLE: struct_field.nullable(),
        }
        comment = struct_field.comment()
        if comment is not None:
            struct_field_data.update({cls.STRUCT_FIELD_COMMENT: comment})
        return struct_field_data

    @classmethod
    def write_list_type(cls, list_type: Types.ListType) -> Dict[str, Any]:
        list_data = {
            cls.TYPE: cls.LIST,
            cls.LIST_ELEMENT_TYPE: cls.write_data_type(list_type.element_type()),
            cls.LIST_ELEMENT_NULLABLE: list_type.element_nullable(),
        }
        return list_data

    @classmethod
    def write_map_type(cls, map_type: Types.MapType) -> Dict[str, Any]:
        map_data = {
            cls.TYPE: cls.MAP,
            cls.MAP_VALUE_NULLABLE: map_type.is_value_nullable(),
            cls.MAP_KEY_TYPE: cls.write_data_type(data_type=map_type.key_type()),
            cls.MAP_VALUE_TYPE: cls.write_data_type(data_type=map_type.value_type()),
        }
        return map_data

    @classmethod
    def write_union_type(cls, union_type: Types.UnionType) -> Dict[str, Any]:
        union_data = {
            cls.TYPE: cls.UNION,
            cls.UNION_TYPES: [
                cls.write_data_type(data_type=type_) for type_ in union_type.types()
            ],
        }
        return union_data

    @classmethod
    def write_external_type(cls, external_type: Types.ExternalType) -> Dict[str, str]:
        external_data = {
            cls.TYPE: cls.EXTERNAL,
            cls.CATALOG_STRING: external_type.catalog_string(),
        }
        return external_data

    @classmethod
    @overload
    def write_unparsed_type(cls, unparsed_type: str) -> Dict[str, str]: ...

    @classmethod
    @overload
    def write_unparsed_type(cls, unparsed_type: Type) -> Dict[str, str]: ...

    @classmethod
    def write_unparsed_type(cls, unparsed_type) -> Dict[str, str]:
        return {
            cls.TYPE: cls.UNPARSED,
            cls.UNPARSED_TYPE: (
                unparsed_type.unparsed_type()
                if isinstance(unparsed_type, Types.UnparsedType)
                else unparsed_type
            ),
        }

    @classmethod
    def read_data_type(cls, type_data: Json) -> Type:
        """Read Gravitino Type from JSON data. Used for Gravitino Type JSON Deserialization.

        Args:
            type_data (Json): The serialized data.

        Returns:
            Type: The Gravitino Type.
        """
        if type_data is None or isinstance(type_data, (dict, str)):
            Precondition.check_argument(
                type_data is not None and len(type_data) > 0,
                f"Cannot parse type from invalid JSON: {type_data}",
            )

        if isinstance(type_data, str):
            return cls.from_primitive_type_string(type_data.lower())

        if isinstance(type_data, dict) and type_data.get(cls.TYPE):
            type_str = type_data[cls.TYPE]
            if cls.STRUCT == type_str:
                return cls.read_struct_type(type_data)
            if cls.LIST == type_str:
                return cls.read_list_type(type_data)
            if cls.MAP == type_str:
                return cls.read_map_type(type_data)
            if cls.UNION == type_str:
                return cls.read_union_type(type_data)
            if cls.UNPARSED == type_str:
                return cls.read_unparsed_type(type_data)
            if cls.EXTERNAL == type_str:
                return cls.read_external_type(type_data)

        return Types.UnparsedType.of(unparsed_type=json.dumps(type_data))

    @classmethod
    def from_primitive_type_string(cls, type_string: str) -> Type:
        type_instance = cls.TYPES.get(type_string)
        if type_instance is not None:
            return type_instance

        decimal_matched = cls.DECIMAL_PATTERN.match(type_string)
        if decimal_matched:
            return Types.DecimalType.of(
                precision=int(decimal_matched.group(1)),
                scale=int(decimal_matched.group(2)),
            )

        fixed_matched = cls.FIXED_PATTERN.match(type_string)
        if fixed_matched:
            return Types.FixedType.of(length=int(fixed_matched.group(1)))

        fixedchar_matched = cls.FIXEDCHAR_PATTERN.match(type_string)
        if fixedchar_matched:
            return Types.FixedCharType.of(length=int(fixedchar_matched.group(1)))

        varchar_matched = cls.VARCHAR_PATTERN.match(type_string)
        if varchar_matched:
            return Types.VarCharType.of(length=int(varchar_matched.group(1)))

        time_matched = cls.TIME_PATTERN.match(type_string)
        if time_matched:
            return Types.TimeType.of(precision=int(time_matched.group(1)))

        timestamp_matched = cls.TIMESTAMP_PATTERN.match(type_string)
        if timestamp_matched:
            return Types.TimestampType.without_time_zone(
                precision=int(timestamp_matched.group(1))
            )

        timestamp_tz_matched = cls.TIMESTAMP_TZ_PATTERN.match(type_string)
        if timestamp_tz_matched:
            return Types.TimestampType.with_time_zone(
                precision=int(timestamp_tz_matched.group(1))
            )

        return Types.UnparsedType.of(type_string)

    @classmethod
    def read_struct_type(cls, struct_data: Dict[str, Any]) -> Types.StructType:
        fields = struct_data.get(cls.FIELDS)
        Precondition.check_argument(
            fields is not None and isinstance(fields, list),
            f"Cannot parse struct fields from non-array: {fields}",
        )
        field_list = [cls.read_struct_field(field) for field in struct_data[cls.FIELDS]]
        return Types.StructType.of(*field_list)

    @classmethod
    def read_struct_field(cls, field_data: Dict[str, Any]) -> Types.StructType.Field:
        Precondition.check_argument(
            field_data is not None and isinstance(field_data, dict),
            f"Cannot parse struct field from invalid JSON: {field_data}",
        )
        Precondition.check_argument(
            field_data.get(cls.STRUCT_FIELD_NAME) is not None,
            f"Cannot parse struct field from missing name: {field_data}",
        )
        Precondition.check_argument(
            field_data.get(cls.TYPE) is not None,
            f"Cannot parse struct field from missing type: {field_data}",
        )

        name = field_data[cls.STRUCT_FIELD_NAME]
        field_type = cls.read_data_type(field_data[cls.TYPE])
        nullable = field_data.get(cls.STRUCT_FIELD_NULLABLE, True)
        comment = field_data.get(cls.STRUCT_FIELD_COMMENT, "")

        return Types.StructType.Field(
            name=name, field_type=field_type, nullable=nullable, comment=comment
        )

    @classmethod
    def read_list_type(cls, list_data: Dict[str, Any]) -> Types.ListType:
        Precondition.check_argument(
            list_data.get(cls.LIST_ELEMENT_TYPE) is not None,
            f"Cannot parse list type from missing element type: {list_data}",
        )
        element_type = cls.read_data_type(list_data[cls.LIST_ELEMENT_TYPE])
        nullable = list_data.get(cls.LIST_ELEMENT_NULLABLE, True)
        return Types.ListType.of(element_type=element_type, element_nullable=nullable)

    @classmethod
    def read_map_type(cls, map_data: Dict[str, Any]) -> Types.MapType:
        Precondition.check_argument(
            map_data.get(cls.MAP_KEY_TYPE) is not None,
            f"Cannot parse map type from missing key type: {map_data}",
        )
        Precondition.check_argument(
            map_data.get(cls.MAP_VALUE_TYPE) is not None,
            f"Cannot parse map type from missing value type: {map_data}",
        )
        key_type = cls.read_data_type(map_data[cls.MAP_KEY_TYPE])
        value_type = cls.read_data_type(map_data[cls.MAP_VALUE_TYPE])
        nullable = map_data.get(cls.MAP_VALUE_NULLABLE, True)
        return Types.MapType.of(key_type, value_type, nullable)

    @classmethod
    def read_union_type(cls, union_data: Dict[str, Any]) -> Types.UnionType:
        Precondition.check_argument(
            union_data.get(cls.UNION_TYPES) is not None,
            f"Cannot parse union type from missing types: {union_data}",
        )
        types = union_data.get(cls.UNION_TYPES)
        Precondition.check_argument(
            types is not None and isinstance(types, list),
            f"Cannot parse union types from non-array: {types}",
        )

        union_types = [cls.read_data_type(type_data) for type_data in types]
        return Types.UnionType.of(*union_types)

    @classmethod
    def read_unparsed_type(cls, data: Dict[str, Any]) -> Types.UnparsedType:
        Precondition.check_argument(
            data.get(cls.UNPARSED_TYPE) is not None,
            f"Cannot parse unparsed type from missing unparsed type: {data}",
        )

        return Types.UnparsedType.of(data[cls.UNPARSED_TYPE])

    @classmethod
    def read_external_type(cls, external_data: Dict[str, Any]) -> Types.ExternalType:
        Precondition.check_argument(
            external_data.get(cls.CATALOG_STRING) is not None,
            f"Cannot parse external type from missing catalogString: {external_data}",
        )
        return Types.ExternalType.of(external_data[cls.CATALOG_STRING])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum


class Strategy(Enum):
    """
    An enum that defines the distribution strategy.

    The following strategies are supported:

    - NONE: No distribution strategy, depends on the underlying system's allocation.
    - HASH: Uses the hash value of the expression to distribute data.
    - RANGE: Uses the specified range of the expression to distribute data.
    - EVEN: Distributes data evenly across partitions.
    """

    NONE = "NONE"
    HASH = "HASH"
    RANGE = "RANGE"
    EVEN = "EVEN"

    @staticmethod
    def get_by_name(name: str) -> "Strategy":
        upper_name = name.upper()
        if upper_name == "NONE":
            return Strategy.NONE
        if upper_name == "HASH":
            return Strategy.HASH
        if upper_name == "RANGE":
            return Strategy.RANGE
        if upper_name in {"EVEN", "RANDOM"}:
            return Strategy.EVEN
        raise ValueError(
            f"Invalid distribution strategy: {name}. Valid values are: {[s.value for s in Strategy]}"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import abstractmethod
from typing import List

from gravitino.api.rel.expressions.distributions.strategy import Strategy
from gravitino.api.rel.expressions.expression import Expression


class Distribution(Expression):
    """
    An interface that defines how data is distributed across partitions.
    """

    @abstractmethod
    def strategy(self) -> Strategy:
        """Return the distribution strategy name."""

    @abstractmethod
    def number(self) -> int:
        """Return The number of buckets/distribution. For example, if the distribution strategy is HASH
        and the number is 10, then the data is distributed across 10 buckets."""

    @abstractmethod
    def expressions(self) -> List[Expression]:
        """Return The expressions passed to the distribution function."""

    def children(self) -> List[Expression]:
        """
        Returns the child expressions.
        """
        return self.expressions()

    def equals(self, other: "Distribution") -> bool:
        """
        Indicates whether some other object is "equal to" this one.

        Args:
            other (Distribution): The reference distribution object with which to compare.

        Returns:
            bool: True if this object is the same as the other; False otherwise.
        """
        if other is None:
            return False

        return (
            self.strategy() == other.strategy()
            and self.number() == other.number()
            and self.expressions() == other.expressions()
        )
